/*
 * s3fs - FUSE-based file system backed by Amazon S3
 *
 * Copyright(C) 2007 Takeshi Nakatani <ggtakec.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include <string>

#include "common.h"

//-------------------------------------------------------------------
// Global variables
//-------------------------------------------------------------------
bool foreground                   = false;
bool nomultipart                  = false;
bool pathrequeststyle             = false;
bool complement_stat              = false;
bool noxmlns                      = false;
std::string program_name;
std::string service_path          = "/";
std::string s3host                = "https://s3.amazonaws.com";
std::string region                = "us-east-1";
std::string cipher_suites;
std::string instance_name;

std::atomic<long long unsigned> num_requests_head_object;
std::atomic<long long unsigned> num_requests_put_object;
std::atomic<long long unsigned> num_requests_get_object;
std::atomic<long long unsigned> num_requests_delete_object;
std::atomic<long long unsigned> num_requests_list_bucket;
std::atomic<long long unsigned> num_requests_mpu_initiate;
std::atomic<long long unsigned> num_requests_mpu_complete;
std::atomic<long long unsigned> num_requests_mpu_abort;
std::atomic<long long unsigned> num_requests_mpu_upload_part;
std::atomic<long long unsigned> num_requests_mpu_copy_part;

/*
* Local variables:
* tab-width: 4
* c-basic-offset: 4
* End:
* vim600: expandtab sw=4 ts=4 fdm=marker
* vim<600: expandtab sw=4 ts=4
*/
